﻿//*****************************************************************************
//
//  By The Dummy Programmer
//  https://www.thedummyprogrammer.com
//
//*****************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Reflection;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;
using TDP.BaseServices.Infrastructure;
using TDP.BaseServices.Infrastructure.Configuration;
using TDP.BaseServices.Infrastructure.DataAccess.SqlClient;
using TDP.BaseServices.Infrastructure.Logging;
using TDP.BaseServices.Infrastructure.Logging.Abstract;
using TDP.BaseServices.Infrastructure.Security;
using TDP.ChatLib.Messages;
using TDP.ChatLib.Messages.Abstract;
using TDP.ChatLib.Messages.Core;
using TDP.ChatLib.Server.Infrastructure;
using TDP.ChatLib.Server.Infrastructure.ChatRepository;
using TDP.ChatLib.Server.Models;
using TDP.PushLib.Server;

namespace TDP.ChatLib.Server
{
    public class ChatServer
    {
        #region "Event definitions"
        public event EventHandler InitStarted;
        protected virtual void OnInitStarted(EventArgs e)
        {
            EventHandler handler = InitStarted;
            if (handler != null)
            {
                foreach (EventHandler singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public delegate void InitEndedDelegate(object sender, InitEndedEventArgs e);
        public event InitEndedDelegate InitEnded;
        protected virtual void OnInitEnded(InitEndedEventArgs e)
        {
            InitEndedDelegate handler = InitEnded;
            if (handler != null)
            {
                foreach (InitEndedDelegate singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public event EventHandler ShutdownStarted;
        protected virtual void OnShutdownStarted(EventArgs e)
        {
            EventHandler handler = ShutdownStarted;
            if (handler != null)
            {
                foreach (EventHandler singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public event EventHandler ShutdownEnded;
        protected virtual void OnShutdownEnded(EventArgs e)
        {
            EventHandler handler = ShutdownEnded;
            if (handler != null)
            {
                foreach (EventHandler singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }
        #endregion
        private static ILogger _Log = new Logger();
        private bool _ServerRunning;
        private Socket _SocketServer;
        private PushServer<int> _PushServer;
        private X509Certificate2 _SslCertificate;

        public ChatServer()
        {
            _Log.Info("ChatServer -> ChatServer() -> started");

            Config.Init();

            _Log.Info("ChatServer -> ChatServer() -> ending");
        }

        private ChatRepositoryManager GetRepositoryInstance()
        {
            ConfigReader CR = new ConfigReader();
            DBAccess DB = new DBAccess(CR);
            return new ChatRepositoryManager(DB);
        }

        private bool ExecRegisterClient(ChatClient client, RegisterClient message)
        {
            ChatRepositoryManager RepManager = GetRepositoryInstance();
            MethodResult<RegisterModel> Result = RepManager.Register(message.EMail, message.Nickname, message.Password);

            RegisterClientResponse Response = new RegisterClientResponse();

            if (Result.ResultCode != 0)
            {
                // Communicate error to the client
                Response.Success = false;
                Response.ErrorCode = Result.ResultCode;

                switch (Result.ResultCode)
                {
                    case -1: Response.Message = "Some information is missing"; break;
                    case -2: Response.Message = "EMail already used"; break;
                    case -3: Response.Message = "NickName already used"; break;
                    default: Response.Message = "Unknown error"; break;
                }

                client.SendObject(Response);
                
                return false;
            }

            // Registration OK, send response to the client along with the authentication token
            Response.Success = true;
            Response.AuthToken = new AuthenticationToken<int>(Result.ObjectResult.ID, DateTime.Now, DateTime.Now.AddMinutes(Config.AuthTokenDurationMinutes), new AsymmetricCryptographyKey(Config.RSAKey));
            client.SendObject(Response);

            return true;
        }

        private bool ExecLoginClient(ChatClient client, LoginClient message)
        {
            ChatRepositoryManager RepManager = GetRepositoryInstance();
            MethodResult<LoginModel> Result = RepManager.Login(message.EMail, message.Password);

            LoginClientResponse Response = new LoginClientResponse();

            if (Result.ResultCode != 0)
            {
                // Comunicate error to the client
                Response.Success = false;
                Response.ErrorCode = Result.ResultCode;

                switch (Result.ResultCode)
                {
                    case -1: Response.Message = "Some information is missing"; break;
                    case -2: Response.Message = "Invalid credentials"; break;
                    default: Response.Message = "Unknown error"; break;
                }

                client.SendObject(Response);
                
                return false;
            }

            // Login OK, send response to the client along with the authentication token
            Response.Success = true;
            Response.Nickname = Result.ObjectResult.Nickname;

            Response.AuthToken = new AuthenticationToken<int>(Result.ObjectResult.ID, DateTime.Now, DateTime.Now.AddMinutes(Config.AuthTokenDurationMinutes), new AsymmetricCryptographyKey(Config.RSAKey));
            client.SendObject(Response);

            return true;
        }

        private bool ExecRenewToken(ChatClient client, RenewToken message)
        {
            RenewTokenResponse Response = new RenewTokenResponse();
            Response.Success = true;
            Response.AuthToken = new AuthenticationToken<int>(message.AuthToken.ClientID, DateTime.Now, DateTime.Now.AddMinutes(Config.AuthTokenDurationMinutes), new AsymmetricCryptographyKey(Config.RSAKey));
            client.SendObject(Response);
            return true;
        }

        private bool ExecSynchronizeClient(ChatClient client, SynchronizeClient message)
        {
            SynchronizeClientResponse Response = new SynchronizeClientResponse();
            if (!message.AuthToken.IsValid(new AsymmetricCryptographyKey(Config.RSAKey), Config.AuthTokenDurationMinutes))
            {
                Response.Success = false;
                Response.Message = "Invalid authentication token.";
                client.SendObject(Response);
                return false;
            }
            
            MethodResult<List<ContactModel>> ResultContacts;
            MethodResult<List<ContactRequestModel>> ResultContactRequests;
            MethodResult<List<MessageModel>> ResultMessages;

            ChatRepositoryManager RepManager = GetRepositoryInstance();

            // Need to synchronize messages first (if requested) to get the correct pending messages count
            // if contact synchronize request is also requested
            if (message.SynchronizeMessages)
            {
                ResultMessages = RepManager.GetUserMessages(message.AuthToken.ClientID, message.RecipientID);
                if (ResultMessages.ResultCode == 0)
                {
                    Response.Messages = ResultMessages.ObjectResult.Select(t => new MessageItem()
                    {
                        ID = t.ID,
                        Message = t.Message,
                        CreationDate = t.CreationDate,
                        SenderID = t.SenderID,
                        SenderEMail = t.SenderEMail,
                        SenderNickname = t.SenderNickname,
                        RecipientID = t.RecipientID,
                        RecipientEMail = t.RecipientEMail,
                        RecipientNickname = t.RecipientNickname
                    }).ToList();
                }
                else
                {
                    Response.Success = false;
                    Response.ErrorCode = ResultMessages.ResultCode;
                    Response.Message = "Error retrieving messages";
                    client.SendObject(Response);
                    return false;
                }
            }

            if (message.SynchronizeContacts)
            {
                ResultContacts = RepManager.GetUserContacts(message.AuthToken.ClientID);
                if (ResultContacts.ResultCode == 0)
                {
                    Response.Contacts = ResultContacts.ObjectResult.Select(t => new ContactItem() { ID = t.UserID, EMail = t.EMail, NickName = t.Nickname, PendingMessages = t.PendingMessages }).ToList();
                }
                else
                {
                    Response.Success = false;
                    Response.ErrorCode = ResultContacts.ResultCode;
                    Response.Message = "Error retrieving contacts";
                    client.SendObject(Response);
                    return false;
                }
            }
            
            if (message.SynchronizeContactRequests)
            {
                ResultContactRequests = RepManager.GetUserContactRequests(message.AuthToken.ClientID);
                if (ResultContactRequests.ResultCode == 0)
                {
                    Response.ContactRequestes = ResultContactRequests.ObjectResult.Select(t => new ContactRequestItem() { ID = t.ID, UserID = t.UserID, EMail = t.EMail, NickName = t.Nickname }).ToList();
                }
                else
                {
                    Response.Success = false;
                    Response.ErrorCode = ResultContactRequests.ResultCode;
                    Response.Message = "Error retrieving contact requests";
                    client.SendObject(Response);
                    return false;
                }
            }
            
            Response.Success = true;
            client.SendObject(Response);

            return true;
        }

        private bool ExecSendChatMessage(ChatClient client, SendChatMessage message)
        {
            ChatRepositoryManager RepManager = GetRepositoryInstance();
            int Result = RepManager.SendMessage(message.AuthToken.ClientID, message.RecipientID, message.ChatMessage);

            SendChatMessageResponse Response = new SendChatMessageResponse();

            if (Result != 0)
            {
                // Comunicate error to the client
                Response.Success = false;
                Response.ErrorCode = Result;

                switch (Result)
                {
                    case -1: Response.Message = "Some information is missing"; break;
                    case -2: Response.Message = "Invalid sender"; break;
                    case -3: Response.Message = "Invalid recipient"; break;
                    default: Response.Message = "Unknown error"; break;
                }

                client.SendObject(Response);
                return false;
            }

            // Registration OK, send response to the client along with the authentication token
            Response.Success = true;
            client.SendObject(Response);

            // Send push notification to the recipient
            _PushServer.SendPushMessage(message.RecipientID, $"SynchronizeMessages:{message.AuthToken.ClientID}");
         
            return true;
        }

        private bool ExecSearchClient(ChatClient client, SearchClient message)
        {
            ChatRepositoryManager RepManager = GetRepositoryInstance();
            MethodResult<List<UserModel>> Result = RepManager.SearchClient(message.AuthToken.ClientID, message.SearchText);

            SearchClientResponse Response = new SearchClientResponse();

            if (Result.ResultCode != 0)
            {
                // Comunicate error to the client
                Response.Success = false;
                Response.ErrorCode = Result.ResultCode;

                switch (Result.ResultCode)
                {
                    case -1: Response.Message = "Some information is missing"; break;
                    default: Response.Message = "Unknown error"; break;
                }

                client.SendObject(Response);
                return false;
            }

            Response.Success = true;
            Response.Contacts = Result.ObjectResult.Select(t => new ContactItem() { ID = t.ID, EMail = t.EMail, NickName = t.Nickname }).ToList();
            client.SendObject(Response);

            return true;
        }

        private bool ExecAddClient(ChatClient client, AddClient message)
        {
            ChatRepositoryManager RepManager = GetRepositoryInstance();
            int Result = RepManager.AddClient(message.AuthToken.ClientID, message.ID);

            AddClientResponse Response = new AddClientResponse();

            if (Result != 0)
            {
                // Comunicate error to the client
                Response.Success = false;
                Response.ErrorCode = Result;

                switch (Result)
                {
                    case -1: Response.Message = "Some information is missing"; break;
                    case -2: Response.Message = "Invalid RequestorID"; break;
                    case -3: Response.Message = "Invalid RecipientID"; break;
                    case -4: Response.Message = "There is alreasy a pending contact request with this contact"; break;
                    case -5: Response.Message = "The recipient is alreasy a contact"; break;
                    default: Response.Message = "Unknown error"; break;
                }

                client.SendObject(Response);
                return false;
            }

            Response.Success = true;
            client.SendObject(Response);

            // Send push notification to the recipient
            _PushServer.SendPushMessage(message.ID, "SynchronizeContactRequests");

            return true;
        }

        private bool ExecConfirmAddClientRequest(ChatClient client, ConfirmAddClientRequest message)
        {
            ChatRepositoryManager RepManager = GetRepositoryInstance();
            int Result = RepManager.ConfirmAddClientRequest(message.AuthToken.ClientID, message.RequestID);

            GenericResponse Response = new GenericResponse();

            if (Result != 0)
            {
                // Comunicate error to the client
                Response.Success = false;
                Response.ErrorCode = Result;

                switch (Result)
                {
                    case -1: Response.Message = "Some information is missing"; break;
                    case -2: Response.Message = "RequestID and RecipientID mismatch"; break;
                    default: Response.Message = "Unknown error"; break;
                }

                client.SendObject(Response);
                return false;
            }

            MethodResult<List<ContactRequestDetailModel>> RequestDetail = RepManager.GetContactRequestDetail(message.RequestID);
            if (RequestDetail.ObjectResult.Count > 0)
            {
                int RequestorID = RequestDetail.ObjectResult[0].RequestorID;

                // Send push notification to the recipient
                _PushServer.SendPushMessage(RequestorID, "SynchronizeContacts");
            }

            Response.Success = true;
            client.SendObject(Response);

            return true;
        }

        private bool ExecRejectAddClientRequest(ChatClient client, RejectAddClientRequest message)
        {
            ChatRepositoryManager RepManager = GetRepositoryInstance();
            int Result = RepManager.RejectAddClientRequest(message.AuthToken.ClientID, message.RequestID);

            GenericResponse Response = new GenericResponse();

            if (Result != 0)
            {
                // Comunicate error to the client
                Response.Success = false;
                Response.ErrorCode = Result;

                switch (Result)
                {
                    case -1: Response.Message = "Some information is missing"; break;
                    case -2: Response.Message = "RequestID and RecipientID mismatch"; break;
                    default: Response.Message = "Unknown error"; break;
                }

                client.SendObject(Response);
                return false;
            }

            Response.Success = true;
            client.SendObject(Response);

            return true;
        }

        private bool ExecRemoveClient(ChatClient client, RemoveClient message)
        {
            ChatRepositoryManager RepManager = GetRepositoryInstance();
            int Result = RepManager.RemoveClient(message.AuthToken.ClientID, message.ID);

            GenericResponse Response = new GenericResponse();

            if (Result != 0)
            {
                // Comunicate error to the client
                Response.Success = false;
                Response.ErrorCode = Result;

                switch (Result)
                {
                    case -1: Response.Message = "Some information is missing"; break;
                    default: Response.Message = "Unknown error"; break;
                }

                client.SendObject(Response);
                return false;
            }

            Response.Success = true;
            client.SendObject(Response);

            _PushServer.SendPushMessage(message.ID, "SynchronizeContacts");

            return true;
        }


        public async Task StartAsync()
        {
            await Task.Run(async () =>
            {
                try
                {
                    _Log.Info("ChatServer -> StartAsync() -> async task started");

                    OnInitStarted(new EventArgs());

                    Stop();

                    _PushServer = new PushServer<int>();
                    await _PushServer.StartAsync();

                    _SocketServer = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                    IPAddress MyIPAddress = IPAddress.Parse(Config.ListenIPAddress);
                    IPEndPoint MyEndPoint = new IPEndPoint(MyIPAddress, Config.ListenPort);
                    _SocketServer.Bind(MyEndPoint);
                    _SocketServer.Listen(Config.ListenBacklog);

                    if (Config.UseSsl)
                    {
                        if (Config.Certificate == string.Empty)
                        {
                            _Log.Info("ChatServer -> StartAsync() -> Missing SSL certificate configuration");
                            return;
                        }

                        _Log.Info($"ChatServer -> StartAsync() -> Loading certificate {Config.Certificate}");

                        _SslCertificate = new X509Certificate2(Config.Certificate, Config.CertificatePassword);
                    }

                    _ServerRunning = true;
                }
                catch (Exception ex)
                {
                    _Log.Error("ChatServer -> StartAsync() -> Error during initialization", ex);
                    OnInitEnded(new InitEndedEventArgs(new Result(-1, false, "Initialization failed")));
                    return;
                }

                Task TaskWaitForClient = WaitForClientsAsync();
                
                OnInitEnded(new InitEndedEventArgs(new Result(0, true, string.Empty)));

                _Log.Info("ChatServer -> StartAsync() -> async task ending");
            });
        }

        public void Stop()
        {
            try
            {
                if (_SocketServer != null)
                {
                    if (_SocketServer.Connected)
                        _SocketServer.Disconnect(true);

                    _SocketServer.Dispose();
                }

                if (_PushServer != null)
                {
                    _PushServer.Stop();
                }
            }
            catch (Exception ex)
            {
                _Log.Error("ChatServer -> Stop() -> Error stopping the server", ex);
            }

            _ServerRunning = false;
        }

        public async Task StopAsync()
        {
            await Task.Run(() =>
            {
                try
                {
                    _Log.Info("ChatServer -> StopAsync() -> async task started");

                    OnShutdownStarted(new EventArgs());

                    Stop();

                    OnShutdownEnded(new EventArgs());

                    _Log.Info("ChatServer -> StopAsync() -> async task ending");
                }
                catch (Exception ex)
                {
                    _Log.Error("ChatServer -> StopAsync() -> Error during stoppage of the server", ex);
                }
            });
        }

        private async Task WaitForClientsAsync()
        {
            await Task.Run(() =>
            {
                _Log.Info("ChatServer -> WaitForClientsAsync() -> async task started");

                while (_ServerRunning)
                {
                    _Log.Info("ChatServer -> WaitForClientsAsync() -> Waiting for connections");

                    ChatClient NewClient = null;
                    try
                    {
                        Socket ClientSocket = _SocketServer.Accept();
                        NewClient = new ChatClient(ClientSocket, _SslCertificate);
                    }
                    catch (Exception ex)
                    {
                        _Log.Error("ChatServer -> WaitForClientsAsync() -> Error while waiting for connections, the server will be stopped", ex);
                        Stop();
                        return;
                    }

                    _Log.Info("ChatServer -> WaitForClientsAsync() -> Connection accepted");

                    Task ManageClient = ManageClientAsync(NewClient);
                }

                _Log.Info("ChatServer -> WaitForClientsAsync() -> async task ending");
            }
           );
        }

        private bool ExecMessage(ChatClient client, object message)
        {
            bool Result = false;
            if (message.GetType().FullName.StartsWith("TDP.ChatLib.Messages.Core."))
            {
                string MethodName = System.Text.RegularExpressions.Regex.Split(message.GetType().FullName, "TDP.ChatLib.Messages.Core.")[1];
                Type ThisType = this.GetType();
                if (MethodName != "LoginClient" && MethodName != "RegisterClient")
                {
                    IAuthenticatedMessage AuthMsg = (IAuthenticatedMessage)message;
                    if (!AuthMsg.AuthToken.IsValid(new AsymmetricCryptographyKey(Config.RSAKey), Config.AuthTokenDurationMinutes))
                    {
                        Type ResponseType = Type.GetType(message.GetType().FullName + "Response, TDP.ChatLib.Messages");
                        IResponseBase Response = (IResponseBase)Activator.CreateInstance(ResponseType);
                        Response.Success = false;
                        Response.ErrorCode = -1000;
                        Response.Message = "Invalid authentication token.";
                        
                        client.SendObject(Response);
                        return false;
                    }
                }

                MethodInfo TheMethod =  ThisType.GetMethod("Exec" + MethodName, BindingFlags.NonPublic | BindingFlags.Instance);
                Result = (bool)TheMethod.Invoke(this, new object[] {client, message } );
            }

            return Result;
        }

        private async Task ManageClientAsync(ChatClient client)
        {
            await Task.Run(() =>
            {
                _Log.Info("ChatServer -> ManageClientAsync() -> async task started");

                object Message = null;
                try
                {
                    Message = client.ReceiveObject();
                    _Log.Info($"ChatServer -> ManageClientAsync() -> message {Message.GetType().ToString()} received");
                }
                catch (Exception ex)
                {
                    _Log.Error("ChatServer -> ManageClientAsync() -> Error receiving message", ex);
                }
                
                if (Message != null)
                {
                    bool Result = false;
                    try
                    {
                        Result = ExecMessage(client, Message);
                        _Log.Info($"ChatServer -> ManageClientAsync() -> message managed with result {Result}, about to close socket connection");
                    }
                    catch (Exception ex)
                    {
                        _Log.Error("ChatServer -> ManageClientAsync() -> Error calling ExecMessage, about to close socket connection", ex);
                    }
                }
                
                try
                {
                    client.Disconnect();
                }
                catch (Exception ex)
                {
                    _Log.Error("ChatServer -> ManageClientAsync() -> Error closing socket connection", ex);
                }

                _Log.Info("ChatServer -> ManageClientAsync() -> async task ending");
            });
        }
    }
}
